from scout.parse.variant.managed_variant import parse_managed_variant_lines


def test_parse_managed_variant_lines(managed_variants_lines):
    """Test parsing managed variant csv"""

    ## GIVEN an iterable with managed variant lines
    nr_mvs = len([line for line in managed_variants_lines if not line.startswith("#")])

    ## WHEN parsing managed variant lines
    mvs = parse_managed_variant_lines(managed_variants_lines)

    # one dict is returned for each line
    assert len(mvs) == nr_mvs
    # and some mv content is parsed to the given key
    assert mvs[0]["chromosome"] == "14"
    assert mvs[0]["position"] == "76548781"
    assert mvs[0]["description"] == "IFT43 indel test"



"""Code for handler that helps export bson to json"""

import datetime
import bson


def bson_handler(field):
    if isinstance(field, datetime.datetime):
        return field.isoformat()
    if isinstance(field, bson.objectid.ObjectId):
        return str(field)
    else:
        raise TypeError(field)



import copy
from werkzeug.datastructures import MultiDict
from scout.server.extensions import store

from scout.server.blueprints.institutes.controllers import cases, phenomodel_checkgroups_filter


def test_phenomodel_checkgroups_filter(app, institute_obj, hpo_checkboxes, omim_checkbox):
    """Test the controllers function that updates the phenotype model based on the model preview checkbox preferences"""
    # GIVEN a database with the required HPO terms (one parent term and one child term)
    store.hpo_term_collection.insert_many(hpo_checkboxes)

    # GIVEN a phenotype model with one panel and 3 checkboxes (checkbox2 nested inside checkbox1)
    hpo_id1 = hpo_checkboxes[0]["_id"]
    hpo_id2 = hpo_checkboxes[1]["_id"]
    omim_id = omim_checkbox["_id"]
    checkbox2 = dict(
        name=hpo_id2,
        description=hpo_checkboxes[1]["description"],
    )
    checkbox1 = dict(
        name=hpo_id1,
        description=hpo_checkboxes[1]["description"],
        children=[checkbox2],  # nested checkbox with HPO term 2
    )
    checkbox3 = {"name": omim_id, "description": omim_checkbox["description"]}
    test_model = dict(
        institute=institute_obj["_id"],
        name="Test model",
        subpanels=dict(panel1=dict(checkboxes={hpo_id1: checkbox1, omim_id: checkbox3})),
    )
    store.phenomodel_collection.insert_one(test_model)
    model_obj = store.phenomodel_collection.find_one()
    # THEN 2 checkboxes should show at the top level
    assert model_obj["subpanels"]["panel1"]["checkboxes"][hpo_id1]
    assert model_obj["subpanels"]["panel1"]["checkboxes"][omim_id]
    assert hpo_id2 not in model_obj["subpanels"]["panel1"]["checkboxes"]

    with app.app_context():
        # WHEN phenotype checkboxes are updated to keep only the nested checkbox2
        checked_terms = MultiDict({"cheked_terms": [".".join(["panel1", hpo_id1, hpo_id2])]})
        updated_model = phenomodel_checkgroups_filter(model_obj, checked_terms)
        # THEN the right checkbox should be present in submodel checkboxes dictionary
        assert updated_model["subpanels"]["panel1"]["checkboxes"][hpo_id2]
        assert hpo_id1 not in updated_model["subpanels"]["panel1"]["checkboxes"]
        assert omim_id not in updated_model["subpanels"]["panel1"]["checkboxes"]


def test_cases(adapter, case_obj, institute_obj):

    # GIVEN a non prioritized case
    case = case_obj
    assert case["status"] == "inactive"
    adapter.case_collection.insert_one(case)

    # GIVEN a priotized case
    case2 = copy.deepcopy(case)
    case2["_id"] = "internal_id2"
    case2["status"] = "prioritized"
    adapter.case_collection.insert_one(case2)

    all_cases = adapter.cases(collaborator=institute_obj["_id"])
    assert len(list(all_cases)) == 2

    prio_cases = adapter.prioritized_cases(institute_id=institute_obj["_id"])
    assert len(list(prio_cases)) == 1

    all_cases = adapter.cases(collaborator=institute_obj["_id"])
    prio_cases = adapter.prioritized_cases(institute_id=institute_obj["_id"])

    # WHEN the cases controller is invoked
    data = cases(store=adapter, case_query=all_cases, prioritized_cases_query=prio_cases, limit=1)

    # THEN 2 cases should be returned
    assert data["found_cases"] == 2


def test_controller_cases(adapter):
    # GIVEN an adapter with a case
    dummy_case = {
        "case_id": "1",
        "owner": "cust000",
        "individuals": [
            {"analysis_type": "wgs", "sex": 1, "phenotype": 2, "individual_id": "ind1"}
        ],
        "status": "inactive",
    }
    adapter.case_collection.insert_one(dummy_case)
    case_query = adapter.case_collection.find()
    # WHEN fetching a case with the controller
    data = cases(adapter, case_query)
    # THEN
    assert isinstance(data, dict)
    assert data["found_cases"] == 1



# igv.js track settings common for all users and all cases
HG19REF_URL = "https://s3.amazonaws.com/igv.broadinstitute.org/genomes/seq/hg19/hg19.fasta"
HG19REF_INDEX_URL = (
    "https://s3.amazonaws.com/igv.broadinstitute.org/genomes/seq/hg19/hg19.fasta.fai"
)
HG19CYTOBAND_URL = "https://s3.amazonaws.com/igv.broadinstitute.org/genomes/seq/hg19/cytoBand.txt"
HG38REF_URL = "https://s3.amazonaws.com/igv.broadinstitute.org/genomes/seq/hg38/hg38.fa"
HG38REF_INDEX_URL = "https://s3.amazonaws.com/igv.broadinstitute.org/genomes/seq/hg38/hg38.fa.fai"
HG38CYTOBAND_URL = (
    "https://s3.amazonaws.com/igv.broadinstitute.org/annotations/hg38/cytoBandIdeo.txt"
)

HG38GENES_FORMAT = "gtf"
HG38GENES_URL = "https://s3.amazonaws.com/igv.broadinstitute.org/annotations/hg38/genes/Homo_sapiens.GRCh38.80.sorted.gtf.gz"
HG38GENES_INDEX_URL = "https://s3.amazonaws.com/igv.broadinstitute.org/annotations/hg38/genes/Homo_sapiens.GRCh38.80.sorted.gtf.gz.tbi"
HG19GENES_FORMAT = "bed"
HG19GENES_URL = (
    "https://s3.amazonaws.com/igv.broadinstitute.org/annotations/hg19/genes/refGene.hg19.bed.gz"
)
HG19GENES_INDEX_URL = (
    "https://s3.amazonaws.com/igv.broadinstitute.org/annotations/hg19/genes/refGene.hg19.bed.gz.tbi"
)

HG38CLINVAR_URL = "https://hgdownload.soe.ucsc.edu/gbdb/hg38/bbi/clinvar/clinvarMain.bb"
HG19CLINVAR_URL = "https://hgdownload.soe.ucsc.edu/gbdb/hg19/bbi/clinvar/clinvarMain.bb"

HG38CLINVAR_CNVS_URL = "https://hgdownload.soe.ucsc.edu/gbdb/hg38/bbi/clinvar/clinvarCnv.bb"
HG19CLINVAR_CNVS_URL = "https://hgdownload.soe.ucsc.edu/gbdb/hg19/bbi/clinvar/clinvarCnv.bb"

# Human genome reference genome build 37. Always displayed
HUMAN_REFERENCE_37 = {
    "fastaURL": HG19REF_URL,
    "indexURL": HG19REF_INDEX_URL,
    "cytobandURL": HG19CYTOBAND_URL,
}

# Human genome reference genome build 38. Always displayed
HUMAN_REFERENCE_38 = {
    "fastaURL": HG38REF_URL,
    "indexURL": HG38REF_INDEX_URL,
    "cytobandURL": HG38CYTOBAND_URL,
}

# igv.js track settings common for all users and all cases. Selectable by users
# Clinvar SNVs track genome build 37
CLINVAR_SNV_37 = {
    "name": "ClinVar",
    "track_name": "clinvar_snvs",
    "type": "annotation",
    "sourceType": "file",
    "displayMode": "EXPANDED",
    "format": "bigBed",
    "maxRows": 30,
    "url": HG19CLINVAR_URL,
}

# Clinvar SNVs track genome build 38
CLINVAR_SNV_38 = {
    "name": "ClinVar",
    "track_name": "clinvar_snvs",
    "type": "annotation",
    "sourceType": "file",
    "displayMode": "EXPANDED",
    "format": "bigBed",
    "maxRows": 30,
    "url": HG38CLINVAR_URL,
}

# ClinVar CNVs track genome build 37
CLINVAR_CNV_37 = {
    "name": "ClinVar CNVs",
    "track_name": "clinvar_cnvs",
    "type": "annotation",
    "sourceType": "file",
    "displayMode": "SQUISHED",
    "format": "bigBed",
    "height": 100,
    "url": HG19CLINVAR_CNVS_URL,
}

# ClinVar CNVs track genome build 38
CLINVAR_CNV_38 = {
    "name": "ClinVar CNVs",
    "track_name": "clinvar_cnvs",
    "type": "annotation",
    "sourceType": "file",
    "displayMode": "SQUISHED",
    "format": "bigBed",
    "height": 100,
    "url": HG38CLINVAR_CNVS_URL,
}

# Human genes track, build 37
HUMAN_GENES_37 = {
    "name": "Genes",
    "track_name": "genes_track",
    "type": "annotation",
    "sourceType": "file",
    "displayMode": "EXPANDED",
    "format": HG19GENES_FORMAT,
    "url": HG19GENES_URL,
    "indexURL": HG19GENES_INDEX_URL,
}

# Human genes track, build 38
HUMAN_GENES_38 = {
    "name": "Genes",
    "track_name": "genes_track",
    "type": "annotation",
    "sourceType": "file",
    "displayMode": "EXPANDED",
    "format": HG38GENES_FORMAT,
    "url": HG38GENES_URL,
    "indexURL": HG38GENES_INDEX_URL,
}

CASE_SPECIFIC_TRACKS = {
    "rhocall_bed": "Rhocall Zygosity",
    "rhocall_wig": "Rhocall Regions",
    "tiddit_coverage_wig": "TIDDIT Coverage",
    "upd_regions_bed": "UPD regions",
    "upd_sites_bed": "UPD sites",
}

HUMAN_REFERENCE = {"37": HUMAN_REFERENCE_37, "38": HUMAN_REFERENCE_38}

USER_DEFAULT_TRACKS = ["Genes", "ClinVar", "ClinVar CNVs"]

# Export selectable custom tracks into lists
IGV_TRACKS = {
    "37": [HUMAN_GENES_37, CLINVAR_SNV_37, CLINVAR_CNV_37],
    "38": [HUMAN_GENES_38, CLINVAR_SNV_38, CLINVAR_CNV_38],
}



# -*- coding: utf-8 -*-
from scout.server.blueprints.alignviewers import controllers


def test_make_igv_tracks():
    """ Test function that creates custom track dictionaries """

    # GIVEN a test track with 2 files
    file_list = ["sample_1_file", "sample_2_file"]
    track_list = controllers.make_igv_tracks("test_track", file_list)

    # The function should return a track list with 2 dictionaries
    assert len(track_list) == 2
    # Containing the expected info
    for track in track_list:
        assert track["name"] == "test_track"
        assert track["url"] in file_list


def test_sample_tracks():
    """ Test the function that creates case individual tracks """

    # GIVEN a case with 3 samples alignments
    sample_names = ["sample1", "sample2", "sample3"]
    sample_bams = ["bam1", "bam2", "bam3"]

    form_data = {
        "sample": ",".join(sample_names),
        "align": "bam",
        "bam": ",".join(sample_bams),
        "bai": "bai1,bai2,bai3",
    }

    display_obj = {}
    # WHEN the set_sample_tracks function is invoked:
    controllers.set_sample_tracks(display_obj, form_data)
    # THEN it should return 3 tracks
    assert len(display_obj["sample_tracks"]) == 3
    # Containing the expected fields
    for track in display_obj["sample_tracks"]:
        assert track["name"] in sample_names
        assert track["url"] in sample_bams


def set_case_specific_tracks():
    """Test function that creates tracks based on case-specific files:
    (rhocall files, tiddit coverage files, upd regions and sites files)
    """
    # GIVEN a case with a rhocall bed file and a TIDDIT wig file
    form_data = {"rhocall_bed": "rhocall.bed", "tiddit_coverage_wig": "tiddit_coverage.wig"}
    # THE case_specific_tracks function should return the expected tracks
    display_obj = {}
    controllers.set_case_specific_tracks(display_obj, form_data)
    assert display_obj["rhocall_bed"]["url"] == form_data["rhocall_bed"]
    assert display_obj["tiddit_coverage_wig"] == form_data["tiddit_coverage_wig"]



# -*- coding: utf-8 -*-

"""
scout.models.phenomodel
~~~~~~~~~~~~~~~~~~
Define a document to describe a phenomodel dictionary

Phenomodels are stored in phenomodel collection
"""
from bson import ObjectId
from datetime import datetime

phenomodel = dict(
    _id=ObjectId,
    institute=str,  # institute["_id"]
    name=str,  # name of phenotype model, example: Epilepsy
    description=str,  # phenomodel description
    subpanels=list,  # list of subpanel dictionaries
    created=datetime,
    updated=datetime,
)

# A subpanel dictionary
subpanel = dict(
    _id=dict(  # id is md5 string
        title=str,  # Subpanel title
        subtitle=str,  # Subpanel subtitle
        created=datetime,
        updated=datetime,
        checkboxes=list,  # a list of checkboxes dictionaries
    )
)

# A checkbox dictionary
checkbox = dict(
    term_id=dict(  # term_id is an HPO term (HP:0002133) or an OMIM term(OMIM:117100)
        description=str,  # HPO term description of OMIM term description ir custom description
        name=str,  # an HPO term (HP:0002133) or an OMIM term(OMIM:117100)
        checkbox_type=str,  # hpo or omim
        children=list,  # present only id checkbox_type=hpo. A list of checkbox dictionaries representing eventual nested HPO terms and their children.
    )
)



# -*- coding: utf-8 -*-
from flask import url_for
from scout.server.extensions import store


def test_managed_variants(app, user_obj, institute_obj):
    """ Test managed variants view """
    # GIVEN an initialized app
    # GIVEN a valid user and institute

    with app.test_client() as client:
        # GIVEN that the user could be logged in
        resp = client.get(url_for("auto_login"))
        assert resp.status_code == 200

        # WHEN accessing the managed variants page
        resp = client.get(url_for("managed_variants.managed_variants"))

        # THEN it should return a page
        assert resp.status_code == 200


def test_add_and_remove_managed_variants(app, user_obj, institute_obj):
    """Test first managed variants views:
    adding a managed variant,
    trying to add it again in duplicate and finally removing it.
    """
    # GIVEN an initialized app
    # GIVEN a user and institute

    with app.test_client() as client:
        # GIVEN that the user could be logged in
        resp = client.get(url_for("auto_login"))
        assert resp.status_code == 200

        # GIVEN data from a filled out form
        add_form_data = {
            "chromosome": "14",
            "position": 76548781,
            "end": 76548781,
            "category": "snv",
            "sub_category": "indel",
            "reference": "CTGGACC",
            "alternative": "G",
        }

        # WHEN attempting to add a valid variant
        resp = client.post(
            url_for("managed_variants.add_managed_variant"),
            data=add_form_data,
        )
        # THEN status should be redirect
        assert resp.status_code == 302
        # THEN one variant should be present in the managed variant collection
        assert sum(1 for i in store.managed_variant_collection.find()) == 1
        # THEN the correct variant should be found inserted into the database
        test_managed_variant = store.managed_variant_collection.find_one()
        assert test_managed_variant["chromosome"] == "14"

        # THEN attempting a duplicate insertion will return an error
        resp = client.post(
            url_for("managed_variants.add_managed_variant"),
            data=add_form_data,
        )
        # THEN the status code should still be redirect
        assert resp.status_code == 302
        # THEN the database should still contatain only one variant
        assert sum(1 for i in store.managed_variant_collection.find()) == 1

        # WHEN requesting to remove the selected variant
        resp = client.post(
            url_for(
                "managed_variants.remove_managed_variant",
                variant_id=test_managed_variant["managed_variant_id"],
            )
        )
        # THEN the status code should again be redirect
        assert resp.status_code == 302
        # THEN no variants should remain in the collection
        assert sum(1 for i in store.managed_variant_collection.find()) == 0



import logging

from scout.models.managed_variant import ManagedVariant

LOG = logging.getLogger(__name__)


def build_managed_variant(managed_variant_info):
    """Build a managed_variant object
    Args:
        managed_variant_info(dict): Compare scout.models.ManagedVariant for details

    Returns:
        managed_variant_obj(scout.models.ManagedVariant): A dictionary with variant information

    """

    end = managed_variant_info.get("end", None)
    if end:
        managed_variant_info["end"] = int(end)

    try:
        managed_variant = ManagedVariant(
            chromosome=str(managed_variant_info["chromosome"]),
            position=int(managed_variant_info["position"]),
            end=managed_variant_info.get("end", None),
            reference=managed_variant_info["reference"],
            alternative=managed_variant_info["alternative"],
            build=managed_variant_info.get("build", "37"),
            maintainer=managed_variant_info.get("", []),
            institute=managed_variant_info.get("institute", ""),
            category=managed_variant_info.get("category", "snv"),
            sub_category=managed_variant_info.get("sub_category", "snv"),
            description=managed_variant_info.get("description", ""),
        )
    except KeyError:
        raise KeyError("Managed variant has to have chr, pos, ref and alt.")

    LOG.debug("Built managed variant %s", managed_variant.get("display_id"))

    return managed_variant



# -*- coding: utf-8 -*-
from scout.commands import cli
from scout.server.extensions import store

test_exon = {
    "exon_id": "1-1167658-1168649",
    "chrom": "1",
    "start": 1167658,
    "end": 1168649,
    "transcript": "ENST00000379198",
    "hgnc_id": 17978,
    "rank": 1,
    "strand": 1,
    "build": "37",
}

test_transcript = {
    "ensembl_transcript_id": "ENST00000379198",
    "hgnc_id": 17978,
    "chrom": "1",
    "start": 1167629,
    "end": 1170421,
    "is_primary": True,
    "refseq_id": "NM_080605",
    "refseq_identifiers": ["NM_080605"],
    "build": "37",
    "length": 2792,
}


def test_export_exons_one_gene(mock_app, gene_obj):
    """Test the CLI command that exports the exons of one gene"""

    runner = mock_app.test_cli_runner()

    # GIVEN a database with a gene
    assert store.hgnc_collection.insert_one(gene_obj)

    # with at least one exon:
    assert store.exon_collection.insert_one(test_exon)

    # WHEN the command to export exons from test gene is invoked
    result = runner.invoke(cli, ["export", "exons", "-b", "37", "-hgnc", gene_obj["hgnc_id"]])
    # it should return the exon
    assert result.exit_code == 0
    assert test_exon["exon_id"] in result.output


def test_export_exons(mock_app, gene_obj):
    """Test the CLI command that exports all exons for a given genome build"""

    runner = mock_app.test_cli_runner()

    # GIVEN a database with a gene
    assert store.hgnc_collection.insert_one(gene_obj)

    # HAVING a transcript
    assert store.transcript_collection.insert_one(test_transcript)

    # AND an exon
    assert store.exon_collection.insert_one(test_exon)

    # WHEN the the command to export all exons is invoked
    result = runner.invoke(cli, ["export", "exons", "-b", "37"])

    # it should return the exon
    assert result.exit_code == 0
    assert test_exon["exon_id"] in result.output


def test_export_exons_gene_json(mock_app, gene_obj):
    """Test the CLI command that exports all exons a gene on json format"""

    runner = mock_app.test_cli_runner()

    # GIVEN a database with a gene
    assert store.hgnc_collection.insert_one(gene_obj)

    # with at least one exon:
    assert store.exon_collection.insert_one(test_exon)

    # WHEN the command to json-export exons from test gene is invoked
    result = runner.invoke(cli, ["export", "exons", "-hgnc", gene_obj["hgnc_id"], "--json"])
    assert result.exit_code == 0
    # THEN it should return a document
    assert "$oid" in result.output
    # With test exon
    assert test_exon["exon_id"] in result.output


def test_export_exons_json(mock_app, gene_obj):
    """Test the CLI command that exports all exons in json format"""

    runner = mock_app.test_cli_runner()

    # GIVEN a database with a gene
    assert store.hgnc_collection.insert_one(gene_obj)

    # HAVING a transcript
    assert store.transcript_collection.insert_one(test_transcript)

    # AND an exon
    assert store.exon_collection.insert_one(test_exon)

    # WHEN the the command to json-export all exons is invoked
    result = runner.invoke(cli, ["export", "exons", "-b", "37", "--json"])
    assert result.exit_code == 0

    # THEN it should return a document
    assert "$oid" in result.output
    # With test exon
    assert test_exon["exon_id"] in result.output



"""Tests for update individual command"""

from click.testing import CliRunner
from scout.commands.update.individual import individual as ind_cmd


def test_update_individual_no_args():
    """Tests the CLI that updates a individual"""

    # GIVEN a CLI object
    runner = CliRunner()

    # WHEN updating a individual without any args
    result = runner.invoke(ind_cmd)

    # THEN assert it fails since its missing mandatory arguments
    assert result.exit_code == 2
    assert "Missing option" in result.output


def test_update_individual_non_existing_case(empty_mock_app):
    """Tests the CLI that updates an individual from a case that doesn't exist"""

    case_id = "acase"
    ind_id = "anind"
    # GIVEN a CLI object
    runner = empty_mock_app.test_cli_runner()

    # WHEN updating a individual without a valid case
    result = runner.invoke(ind_cmd, ["--case-id", case_id, "--ind", ind_id])

    # THEN assert it exits
    assert result.exit_code == 0
    # With a pertinent message
    assert "Could not find case" in result.output


def test_update_individual_no_individual(empty_mock_app, real_adapter, case_obj):
    """Tests the CLI that updates a individual with case id and no ind name"""

    # GIVEN a CLI object
    runner = empty_mock_app.test_cli_runner()

    real_adapter.case_collection.insert_one(case_obj)
    assert real_adapter.case_collection.find_one()

    # WHEN updating a individual providing the wrong individual name or no individual name
    result = runner.invoke(ind_cmd, ["--case-id", case_obj["_id"]])

    # THEN assert it exits
    assert result.exit_code == 0
    # With a pertinent message
    assert "Please specify individual name" in result.output


def test_update_individual_no_update_key(empty_mock_app, real_adapter, case_obj):
    """Tests the CLI that updates a individual without providing a valid key"""

    # GIVEN a CLI object
    runner = empty_mock_app.test_cli_runner()

    real_adapter.case_collection.insert_one(case_obj)
    assert real_adapter.case_collection.find_one()

    ind_name = case_obj["individuals"][0]["display_name"]

    # WHEN updating a individual without a valid key
    result = runner.invoke(ind_cmd, ["--case-id", case_obj["_id"], "--ind", ind_name])

    # THEN assert it exits
    assert result.exit_code == 0
    # With a pertinent message
    assert "Please specify a valid key" in result.output


def test_update_key_non_existent_path(mock_app, real_populated_database, case_obj):
    """Tests the CLI that updates a individual with an alignment path that is nit found on the server"""

    # GIVEN a CLI object
    runner = mock_app.test_cli_runner()
    # GIVEN a database with a case_obj
    existing_case = real_populated_database.case_collection.find_one()
    case_id = existing_case["_id"]
    ind_info = existing_case["individuals"][0]
    ind_name = ind_info["display_name"]

    # WHEN updating a individual without a valid alignment path
    result = runner.invoke(
        ind_cmd,
        ["--case-id", case_id, "--ind", ind_name, "bam_file", "not_a_valid_path"],
    )

    # THEN the command should ask user to confirm save action
    assert result.exit_code == 1
    assert "The provided path was not found on the server, update key anyway?" in result.output


def test_update_alignment_path(mock_app, real_populated_database, bam_path):
    """Tests the CLI that updates a individual with an alignment path"""

    # GIVEN a CLI object
    runner = mock_app.test_cli_runner()
    # GIVEN a database with a case_obj
    existing_case = real_populated_database.case_collection.find_one()
    case_id = existing_case["_id"]
    ind_info = existing_case["individuals"][0]
    ind_name = ind_info["display_name"]

    # WHEN updating a individual with a valid alignment path
    result = runner.invoke(
        ind_cmd,
        ["--case-id", case_id, "--ind", ind_name, "bam_file", str(bam_path)],
    )

    # THEN the command should run with no errors
    assert result.exit_code == 0
    # THEN assert that the new alignment path was added
    fetched_case = real_populated_database.case_collection.find_one()
    assert fetched_case["individuals"][0]["bam_file"] == str(bam_path)



# -*- coding: utf-8 -*-
import logging
import operator
from pymongo.errors import DuplicateKeyError, BulkWriteError
import pymongo
from anytree import RenderTree, Node, search, resolver
from anytree.exporter import DictExporter

from scout.exceptions import IntegrityError

LOG = logging.getLogger(__name__)


class HpoHandler(object):
    def load_hpo_term(self, hpo_obj):
        """Add a hpo object

        Arguments:
            hpo_obj(dict)

        """
        LOG.debug("Loading hpo term %s into database", hpo_obj["_id"])
        try:
            self.hpo_term_collection.insert_one(hpo_obj)
        except DuplicateKeyError as err:
            raise IntegrityError("Hpo term %s already exists in database".format(hpo_obj["_id"]))
        LOG.debug("Hpo term saved")

    def load_hpo_bulk(self, hpo_bulk):
        """Add a hpo object

        Arguments:
            hpo_bulk(list(scout.models.HpoTerm))

        Returns:
            result: pymongo bulkwrite result

        """
        LOG.debug("Loading hpo bulk")

        try:
            result = self.hpo_term_collection.insert_many(hpo_bulk)
        except (DuplicateKeyError, BulkWriteError) as err:
            raise IntegrityError(err)
        return result

    def hpo_term(self, hpo_id):
        """Fetch a hpo term

        Args:
            hpo_id(str)

        Returns:
            hpo_obj(dict)
        """
        LOG.debug("Fetching hpo term %s", hpo_id)

        return self.hpo_term_collection.find_one({"_id": hpo_id})

    def hpo_terms(self, query=None, hpo_term=None, text=None, limit=None):
        """Return all HPO terms

        If a query is sent hpo_terms will try to match with regex on term or
        description.

        Args:
            query(str): Part of a hpoterm or description
            hpo_term(str): Search for a specific hpo term
            limit(int): the number of desired results

        Returns:
            result(pymongo.Cursor): A cursor with hpo terms
        """
        query_dict = {}
        search_term = None
        if query:
            query_dict = {
                "$or": [
                    {"hpo_id": {"$regex": query, "$options": "i"}},
                    {"description": {"$regex": query, "$options": "i"}},
                ]
            }
            search_term = query
        elif text:
            new_string = ""
            for i, word in enumerate(text.split(" ")):
                if i == 0:
                    new_string += word
                else:
                    new_string += ' "{0}"'.format(word)
            LOG.info("Search HPO terms with %s", new_string)
            query_dict["$text"] = {"$search": new_string}
            search_term = text
        elif hpo_term:
            query_dict["hpo_id"] = hpo_term
            search_term = hpo_term

        limit = limit or 0
        res = (
            self.hpo_term_collection.find(query_dict)
            .limit(limit)
            .sort("hpo_number", pymongo.ASCENDING)
        )

        return res

    def generate_hpo_gene_list(self, *hpo_terms):
        """Generate a sorted list with namedtuples of hpogenes
        Each namedtuple of the list looks like (hgnc_id, count)
        Args:
            hpo_terms(iterable(str))
        Returns:
            hpo_genes(list(HpoGene))
        """
        genes = {}
        for term in hpo_terms:
            hpo_obj = self.hpo_term(term)
            if hpo_obj:
                for hgnc_id in hpo_obj["genes"]:
                    if hgnc_id in genes:
                        genes[hgnc_id] += 1
                    else:
                        genes[hgnc_id] = 1
            else:
                LOG.warning("Term %s could not be found", term)

        sorted_genes = sorted(genes.items(), key=operator.itemgetter(1), reverse=True)
        return sorted_genes

    def organize_tree(self, all_terms, root):
        """Organizes a set of Tree node objects into a tree, according to their ancestors and children

        Args:
            all_terms(dict): a dictionary with "term_name" as keys and term_dict as values
            root(anytree.Node)
        Returns
            root(anytree.Node): the updated root node of the tree
        """
        # Move tree nodes in the right position according to the ontology
        for key, term in all_terms.items():
            ancestors = term["ancestors"]
            if len(ancestors) == 0:
                continue
            for ancestor in ancestors:
                ancestor_node = search.find_by_attr(root, ancestor)
                if ancestor_node is None:  # It's probably the term on the top
                    continue
                node = search.find_by_attr(root, key)
                node.parent = ancestor_node
        return root

    def build_phenotype_tree(self, hpo_id):
        """Creates an HPO Tree based on one or more given ancestors
        Args:
            hpo_id(str): an HPO term
        Returns:
            tree_dict(dict): a tree of all HPO children of the given term, as a dictionary
        """
        root = Node(id="root", name="root", parent=None)
        all_terms = {}
        unique_terms = set()

        def _hpo_terms_list(hpo_ids):
            for term_id in hpo_ids:
                term_obj = self.hpo_term(term_id)
                if term_obj is None:
                    continue
                # sort term children by ascending HPO number
                children = sorted(
                    term_obj["children"], key=lambda x: int("".join([i for i in x if i.isdigit()]))
                )
                term_obj["children"] = children
                all_terms[term_id] = term_obj
                if term_id not in unique_terms:
                    node = Node(term_id, parent=root, description=term_obj["description"])
                    unique_terms.add(term_id)
                # recursive loop to collect children, children of children and so on
                _hpo_terms_list(term_obj["children"])

        # compile a list of all HPO term objects to include in the submodel
        _hpo_terms_list([hpo_id])  # trigger the recursive loop to collect nested HPO terms
        # rearrange tree according to the HPO ontology
        root = self.organize_tree(all_terms, root)
        node_resolver = resolver.Resolver("name")
        # Extract a tree structure having the chosen HPO term (hpo_id) as ancestor of all the children terms
        term_node = node_resolver.get(root, hpo_id)
        LOG.info(f"Built ontology for HPO term:{hpo_id}:\n{RenderTree(term_node)}")
        exporter = DictExporter()
        # Export this tree structure as dictionary, so that can be saved in database
        tree_dict = exporter.export(term_node)
        return tree_dict



"""Code for updating reports"""
import logging

import click
from flask.cli import with_appcontext

from scout.load.report import load_delivery_report, load_cnv_report
from scout.server.extensions import store

LOG = logging.getLogger(__name__)


@click.command("delivery-report")
@click.argument("case_id")
@click.argument("report_path", type=click.Path(exists=True))
@click.option("-update", "--update", is_flag=True, help="update delivery report for a sample")
@with_appcontext
def delivery_report(case_id, report_path, update):
    """Add delivery report to an existing case."""

    adapter = store

    try:
        load_delivery_report(
            adapter=adapter,
            case_id=case_id,
            report_path=report_path,
            update=update,
        )
        LOG.info("saved report to case!")
    except Exception as err:
        LOG.error(err)
        raise click.Abort()


@click.command("cnv-report")
@click.argument("case-id", required=True)
@click.argument("report-path", type=click.Path(exists=True), required=True)
@click.option("-u", "--update", is_flag=True, help="update CNV report for a sample")
@with_appcontext
def cnv_report(case_id, report_path, update):
    """Add CNV report report to an existing case."""

    adapter = store

    try:
        load_cnv_report(
            adapter=adapter,
            case_id=case_id,
            report_path=report_path,
            update=update,
        )
        LOG.info("saved report to case!")
    except Exception as err:
        LOG.error(err)
        raise click.Abort()



# -*- coding: utf-8 -*-
import logging

from flask_wtf import FlaskForm
from wtforms import (
    IntegerField,
    SelectField,
    SelectMultipleField,
    SubmitField,
    TextField,
    validators,
)

from scout.constants import (
    SV_TYPES,
    CHROMOSOMES,
)

LOG = logging.getLogger(__name__)
CHROMOSOME_FILTER_OPTIONS = [("", "All")] + [(chrom, chrom) for chrom in CHROMOSOMES]
CHROMOSOME_EDIT_OPTIONS = [(chrom, chrom) for chrom in CHROMOSOMES]
SUBCATEGORY_CHOICES = [("snv", "SNV"), ("indel", "INDEL")] + [
    (term, term.replace("_", " ").upper()) for term in SV_TYPES
]
CATEGORY_CHOICES = [
    (term, term.replace("_", " ").upper()) for term in ["snv", "sv", "cancer", "cancer_sv"]
]


class ManagedVariantForm(FlaskForm):
    position = IntegerField("Start position", [validators.Optional()])
    end = IntegerField("End position", [validators.Optional()])
    cytoband_start = SelectField("Cytoband start", choices=[])
    cytoband_end = SelectField("Cytoband end", choices=[])

    description = TextField(label="Description")


class ManagedVariantsFilterForm(ManagedVariantForm):
    chromosome = SelectField(
        "Chromosome", [validators.Optional()], choices=CHROMOSOME_FILTER_OPTIONS
    )

    category = SelectMultipleField("Category", choices=CATEGORY_CHOICES)
    sub_category = SelectMultipleField("Kind", choices=SUBCATEGORY_CHOICES)

    filter_variants = SubmitField(label="Filter variants")
    export = SubmitField(label="Filter and export")


class ManagedVariantEditForm(ManagedVariantForm):
    chromosome = SelectField("Chromosome", [validators.Optional()], choices=CHROMOSOME_EDIT_OPTIONS)

    reference = TextField(label="Ref")
    alternative = TextField(label="Alt")

    category = SelectField("Category", choices=CATEGORY_CHOICES)
    sub_category = SelectField("Kind", choices=SUBCATEGORY_CHOICES)


class ManagedVariantAddForm(ManagedVariantEditForm):
    add_variant = SubmitField(label="Add")
    cancel = SubmitField(label="Cancel")


class ManagedVariantModifyForm(ManagedVariantEditForm):
    modify_variant = SubmitField(label="Save")
    cancel = SubmitField(label="Cancel")



import logging
from datetime import datetime

from pymongo.errors import DuplicateKeyError

from bson import ObjectId
import pymongo

from scout.exceptions import IntegrityError

LOG = logging.getLogger(__name__)


class ManagedVariantHandler(object):
    """Class to handle variant events for the mongo adapter"""

    def load_managed_variant(self, managed_variant_obj):
        """Load a managed variant object

        Args:
            managed_variant_obj(ManagedVariant)

        Returns:
            inserted_id
        """
        try:
            result = self.managed_variant_collection.insert_one(managed_variant_obj)
        except DuplicateKeyError as err:
            raise IntegrityError(
                "Variant %s already exists in database",
                managed_variant_obj["display_id"],
            )

        return result.inserted_id

    def upsert_managed_variant(self, managed_variant_obj, original_obj_id=None):
        """Load or updated a managed variant object

        Args:
            managed_variant_obj(ManagedVariant)
            original_obj_id(str):   String representation of original obj id to update

        Returns:
            updated_managed_variant
        """

        LOG.debug("Upserting variant %s", managed_variant_obj["display_id"])

        managed_variant_obj["date"] = managed_variant_obj.get("date", datetime.now())

        collision = self.managed_variant_collection.find_one(
            {"managed_variant_id": managed_variant_obj["managed_variant_id"]}
        )
        if collision:
            LOG.debug("Collision - new variant already exists! Leaving variant unmodified.")
            return

        if original_obj_id:
            result = self.managed_variant_collection.find_one_and_update(
                {"_id": ObjectId(original_obj_id)},
                {"$set": managed_variant_obj},
            )

            return result

        try:
            result = self.managed_variant_collection.insert_one(managed_variant_obj)
        except DuplicateKeyError as err:
            LOG.debug(
                "Variant %s already exists in database with a document id %s.",
                managed_variant_obj["display_id"],
                managed_variant_obj["_id"],
            )

        return result

    def managed_variant(self, document_id):
        """Retrieve a managed variant of known id.

        Arguments:
            document_id(ObjectId)

        Returns:
            ManagedVariant
        """

        managed_variant_obj = self.managed_variant_collection.find_one(
            {"_id": ObjectId(document_id)}
        )

        return managed_variant_obj

    def find_managed_variant(self, managed_variant_id):
        """Fetch eg search for a managed variant.

        Arguments:
            display_id(str): chrom_pos_ref_alt_category_build
                category: "snv", "cancer" - "sv", "cancer_sv" possible but not expected
                build: "37" or "38"

        Returns:
            ManagedVariant
        """
        managed_variant = self.managed_variant_collection.find_one(
            {"managed_variant_id": managed_variant_id}
        )

        return managed_variant

    def find_managed_variant_id(self, variant_id):
        """Fetch eg search for a managed variant with the encoded positional variant_id.

        Arguments:
            display_id(str): chrom_pos_ref_alt_category_build
                category: "snv", "cancer" - "sv", "cancer_sv" possible but not expected
                build: "37" or "38"

        Returns:
            ManagedVariant
        """
        managed_variant = self.managed_variant_collection.find_one({"variant_id": variant_id})

        return managed_variant

    def managed_variants(self, category="snv", build="37", query_options=None):
        """Return a cursor to all managed variants of a particular category and build.

        Arguments:
            category(str):
            sub_category(str):
            build(str):

        Returns:
            managed_variants(pymongo.Cursor)

        """

        query = {"category": category, "build": build}
        query_with_options = self.add_options(query, query_options)

        return self.managed_variant_collection.find(query_with_options)

    def count_managed_variants(self, category="snv", build="37", query_options=None):
        """Return count of documents to all managed variants of a particular category and build.

        Arguments:
            category(str):
            sub_category(str):
            build(str):

        Returns:
            integer

        """
        query = {"category": category, "build": build}
        query_with_options = self.add_options(query, query_options)

        return self.managed_variant_collection.count_documents(query_with_options)

    def add_options(self, query, query_options):
        """Update query with `query_options`"""
        # TODO: why is `query_options` both popped and then updated?
        if query_options:
            if "description" in query_options:
                query["description"] = {"$regex": ".*" + query_options["description"] + ".*"}
                query_options.pop("description")

            if "position" in query_options:
                query["end"] = {"$gte": int(query_options["position"])}
                query_options.pop("position")

            if "end" in query_options:
                query["position"] = {"$lte": int(query_options["end"])}
                query_options.pop("end")
            return query.update(query_options)
        return query

    def get_managed_variants(self, institute_id=None, category="snv", build="37"):
        """Return managed variant_ids. Limit by institute, category and build.

        Returns:
            managed_variant_ids(iterable(variant_id: String))
        """

        return [
            managed_variant["variant_id"]
            for managed_variant in self.managed_variants(category, build)
        ]

    def delete_managed_variant(self, managed_variant_id):
        """Delete a managed variant of known id.

        Arguments:
            variant_id(str)

        Returns:
            ManagedVariant
        """

        LOG.info("Attempting to delete managed variant with id %s.", managed_variant_id)

        result = self.managed_variant_collection.find_one_and_delete(
            {"managed_variant_id": managed_variant_id}
        )

        if not result:
            LOG.info(
                "FAILED deleting managed variant: variant_id %s not found.", managed_variant_id
            )

        return result



# -*- coding: utf-8 -*-
import datetime
import pymongo
from bson import ObjectId


class PhenoModelHandler(object):
    """Class handling phenomodels creation and use"""

    def phenomodels(self, institute_id):
        """Return all phenopanels for a given institute
        Args:
            institute_id(str): institute id
        Returns:
            phenotype_models(pymongo.cursor.Cursor)
        """
        query = {"institute": institute_id}
        phenotype_models = self.phenomodel_collection.find(query)
        return phenotype_models

    def create_phenomodel(self, institute_id, name, description):
        """Create an empty advanced phenotype model with data provided by a user
        Args:
            institute_id(str): institute id
            name(str) a model name
            description(str) a model description
        Returns:
            phenomodel_obj(dict) a newly created model
        """
        phenomodel_obj = dict(
            institute=institute_id,
            name=name,
            description=description,
            subpanels={},
            created=datetime.datetime.now(),
            updated=datetime.datetime.now(),
        )
        phenomodel_obj = self.phenomodel_collection.insert_one(phenomodel_obj)
        return phenomodel_obj

    def update_phenomodel(self, model_id, model_obj):
        """Update a phenotype model using its ObjectId
        Args:
            model_id(str): document ObjectId string id
            model_obj(dict): a dictionary of key/values to update a phenomodel with

        Returns:
            updated_model(dict): the phenomodel document after the update
        """
        updated_model = self.phenomodel_collection.find_one_and_update(
            {"_id": ObjectId(model_id)},
            {
                "$set": {
                    "name": model_obj["name"],
                    "description": model_obj["description"],
                    "subpanels": model_obj.get("subpanels", {}),
                    "updated": datetime.datetime.now(),
                }
            },
            return_document=pymongo.ReturnDocument.AFTER,
        )
        return updated_model

    def phenomodel(self, model_id):
        """Retrieve a phenomodel object using its ObjectId
        Args:
            model_id(ObjectId): document ObjectId
        Returns
            model_obj(dict)
        """
        query = {"_id": ObjectId(model_id)}
        model_obj = self.phenomodel_collection.find_one(query)
        return model_obj



import re

from scout.constants import AMINO_ACID_RESIDUE_3_TO_1


def isfloat(x):
    try:
        a = float(x)
    except ValueError:
        return False
    else:
        return True


def isint(x):
    try:
        a = float(x)
        b = int(a)
    except ValueError:
        return False
    else:
        return a == b


def make_bool_pass_none(string):
    if str(string).lower() == "none":
        return None
    return make_bool(string)


def make_bool(string):
    """Convert a string to boolean"""
    if str(string).lower() in ["yes", "true", "1", "t"]:
        return True
    return False


def convert_number(string):
    """Convert a string to number
    If int convert to int otherwise float

    If not possible return None
    """
    res = None
    if isint(string):
        res = int(float(string))
    elif isfloat(string):
        res = float(string)
    return res


def amino_acid_residue_change_3_to_1(protein_sequence_name):
    p = re.compile("p.([A-Za-z]+)(\d+)([A-Za-z]+)")
    m = p.match(protein_sequence_name)
    ref = AMINO_ACID_RESIDUE_3_TO_1.get(m.group(1), None)
    alt = AMINO_ACID_RESIDUE_3_TO_1.get(m.group(3), None)
    pos = m.group(2)
    protein_change = "".join([ref, pos, alt])
    return protein_change



# -*- coding: utf-8 -*-
import logging

from scout.adapter import MongoAdapter
from scout.exceptions import IntegrityError, DataNotFoundError

LOG = logging.getLogger(__name__)


def load_delivery_report(
    adapter: MongoAdapter, report_path: str, case_id: str, update: bool = False
):
    """Load a delivery report into a case in the database

    If the report already exists the function will exit.
    If the user want to load a report that is already in the database
    'update' has to be 'True'

    Args:
        adapter     (MongoAdapter): Connection to the database
        report_path (string):       Path to delivery report
        case_id     (string):       Optional case identifier
        update      (bool):         If an existing report should be replaced

    Returns:
        updated_case(dict)

    """

    case_obj = adapter.case(case_id=case_id)

    if case_obj is None:
        raise DataNotFoundError("no case found")

    if update or case_obj.get("delivery_report") is None:
        _update_report_path(case_obj, report_path, "delivery_report")
    else:
        raise IntegrityError("Existing report found, use update = True to " "overwrite")

    LOG.info("Saving report for case {} in database".format(case_obj["_id"]))
    return adapter.replace_case(case_obj)


def load_cnv_report(adapter: MongoAdapter, report_path: str, case_id: str, update: bool = False):
    """Load a CNV report into a case in the database

    If the report already exists the function will exit.
    If the user want to load a report that is already in the database
    'update' has to be 'True'

    Args:
        adapter     (MongoAdapter): Connection to the database
        report_path (string):       Path to CNV report
        case_id     (string):       Optional case identifier
        update      (bool):         If an existing report should be replaced

    Returns:
        updated_case(dict)

    """

    case_obj = adapter.case(case_id=case_id)

    if case_obj is None:
        raise DataNotFoundError("no case found")

    if update or case_obj.get("cnv_report") is None:
        _update_report_path(case_obj, report_path, "cnv_report")
    else:
        raise IntegrityError("Existing report found, use update = True to " "overwrite")

    LOG.info("Saving report for case {} in database".format(case_obj["_id"]))
    return adapter.replace_case(case_obj)


def _update_report_path(case_obj, report_path, report_type):
    """Updates the report path

    Args:
        case_obj     (Case):         Case object
        report_path  (string):       Path to CNV report
        report_type  (string):       Type of report
    """
    case_obj[report_type] = report_path
    return True



""" Managed variant

    For potentially causative variants that are not yet in ClinVar
    and have yet not been marked causative in any existing case.

"""

from datetime import datetime
from scout.utils.md5 import generate_md5_key


class ManagedVariant(dict):
    """
    # required primary fields
    chromosome=str,  # required
    position=int,  # required
    end=int,  # required
    reference=str,  # required
    alternative=str,  # required
    build=str, # required, ["37","38"], default "37"
    date=datetime.datetime
    # required derived fields
    # display name is variant_id (no md5) chrom_pos_ref_alt (simple_id)
    display_name=str,  # required

    #optional fields
    # maintainer user_id list
    maintainer=list(user_id), # optional
    institute=institute_id, # optional

    # optional fields foreseen for future use
    category=str,  # choices=('sv', 'snv', 'str', 'cancer', 'cancer_sv')
    sub_category=str,  # choices=('snv', 'indel', 'del', 'ins', 'dup', 'inv', 'cnv', 'bnd', 'str')
    description=str,
    """

    def __init__(
        self,
        chromosome,
        position,
        end,
        reference,
        alternative,
        institute,
        maintainer=[],
        build="37",
        date=None,
        category="snv",
        sub_category="snv",
        description=None,
    ):
        super(ManagedVariant, self).__init__()
        self["chromosome"] = str(chromosome)
        self["position"] = position
        self["end"] = end
        self["reference"] = reference
        self["alternative"] = alternative
        self["build"] = build

        self["managed_variant_id"] = "_".join(
            [
                str(part)
                for part in (
                    chromosome,
                    position,
                    reference,
                    alternative,
                    category,
                    sub_category,
                    build,
                )
            ]
        )
        self["display_id"] = "_".join(
            [str(part) for part in (chromosome, position, reference, alternative)]
        )
        self["variant_id"] = generate_md5_key(
            [str(part) for part in (chromosome, position, reference, alternative, "clinical")]
        )
        self["date"] = date or datetime.now()

        self["institute"] = institute or None
        self["maintainer"] = maintainer or []
        self["category"] = category
        self["sub_category"] = sub_category
        self["description"] = description



import logging
import datetime

from flask import Blueprint, request, redirect, url_for, flash
from flask_login import current_user

from scout.server.extensions import store
from scout.server.utils import templated, user_institutes
from .forms import ManagedVariantsFilterForm, ManagedVariantAddForm, ManagedVariantModifyForm
from . import controllers

LOG = logging.getLogger(__name__)

managed_variants_bp = Blueprint("managed_variants", __name__, template_folder="templates")


@managed_variants_bp.route("/managed_variant", methods=["GET", "POST"])
@templated("managed_variants/managed_variants.html")
def managed_variants():
    page = int(request.form.get("page", 1))

    institutes = list(user_institutes(store, current_user))

    filters_form = ManagedVariantsFilterForm(request.form)
    add_form = ManagedVariantAddForm()
    modify_form = ManagedVariantModifyForm()

    category = request.form.get("category", "snv")

    query_options = {}
    for option in ["chromosome", "sub_category", "position", "end", "description"]:
        if request.form.get(option, None):
            query_options[option] = request.form.get(option)

    managed_variants_query = store.managed_variants(category=category, query_options=query_options)
    variant_count = store.count_managed_variants(category=category, query_options=query_options)
    data = controllers.managed_variants(store, managed_variants_query, variant_count, page)

    return dict(
        filters_form=filters_form,
        add_form=add_form,
        modify_form=modify_form,
        page=page,
        **data,
    )


@managed_variants_bp.route("/upload_csv", methods=["POST"])
def upload_managed_variants():
    institutes = list(user_institutes(store, current_user))

    csv_file = request.files["csv_file"]
    content = csv_file.stream.read()
    lines = None
    try:
        if b"\n" in content:
            lines = content.decode("utf-8-sig", "ignore").split("\n")
        else:
            lines = content.decode("windows-1252").split("\r")
    except Exception as err:
        flash(
            "Something went wrong while parsing the panel CSV file! ({})".format(err),
            "danger",
        )
        return redirect(request.referrer)

    LOG.debug("Loading lines %s", lines)
    result = controllers.upload_managed_variants(store, lines, institutes, current_user._id)
    flash(
        "In total {} new variants out of {} in file added".format(result[0], result[1]), "success"
    )

    return redirect(request.referrer)


@managed_variants_bp.route("/managed_variant/<variant_id>/modify", methods=["POST"])
def modify_managed_variant(variant_id):

    edit_form = ManagedVariantModifyForm(request.form)

    if not controllers.modify_managed_variant(store, variant_id, edit_form):
        flash("Could not modify variant - does the new variant perhaps already exist?", "warning")

    return redirect(request.referrer)


@managed_variants_bp.route("/managed_variant/<variant_id>/remove", methods=["POST"])
def remove_managed_variant(variant_id):
    controllers.remove_managed_variant(store, variant_id)

    return redirect(request.referrer)


@managed_variants_bp.route("/managed_variant/add", methods=["POST"])
def add_managed_variant():

    add_form = ManagedVariantAddForm(request.form)
    LOG.debug("Adding managed variant with form %s", add_form)

    institutes = list(user_institutes(store, current_user))
    current_user_id = current_user._id

    controllers.add_managed_variant(store, add_form, institutes, current_user_id)

    return redirect(request.referrer)



# -*- coding: utf-8 -*-
import logging
import os.path

from flask import (
    abort,
    Blueprint,
    render_template,
    send_file,
    request,
    current_app,
    flash,
)

from .partial import send_file_partial
from scout.constants import HUMAN_REFERENCE
from . import controllers

alignviewers_bp = Blueprint(
    "alignviewers",
    __name__,
    template_folder="templates",
    static_folder="static",
    static_url_path="/alignviewers/static",
)

LOG = logging.getLogger(__name__)


@alignviewers_bp.route("/remote/static", methods=["OPTIONS", "GET"])
def remote_static():
    """Stream *large* static files with special requirements."""
    file_path = request.args.get("file")

    range_header = request.headers.get("Range", None)
    if not range_header and (file_path.endswith(".bam") or file_path.endswith(".cram")):
        return abort(500)

    new_resp = send_file_partial(file_path)
    return new_resp


@alignviewers_bp.route("/remote/static/unindexed", methods=["OPTIONS", "GET"])
def unindexed_remote_static():
    file_path = request.args.get("file")
    base_name = os.path.basename(file_path)
    resp = send_file(file_path, attachment_filename=base_name)
    return resp


@alignviewers_bp.route("/igv", methods=["POST"])
def igv():
    """Visualize BAM alignments using igv.js (https://github.com/igvteam/igv.js)"""

    # Set genome build for displaying alignments:
    # Genome build is 37 if request.form.get("build") is 37 and chr != MT
    # Genome build is 38 if request.form.get("build") is 38 or if chrom == MT
    chromosome_build = request.form.get("build")
    chrom = request.form.get("contig")
    if chrom == "MT":
        chrom = "M"
    if chromosome_build in ["GRCh38", "38"] or chrom == "M":
        chromosome_build = "38"
    else:
        chromosome_build = "37"

    start = request.form.get("start")
    stop = request.form.get("stop")
    locus = "chr{0}:{1}-{2}".format(chrom, start, stop)

    display_obj = {}  # Initialize the dictionary containing all tracks info

    # Set up IGV tracks that are common for all cases:
    display_obj["reference_track"] = HUMAN_REFERENCE[
        chromosome_build
    ]  # Human reference is always present
    # General tracks (Genes, Clinvar and ClinVar SNVs are shown according to user preferences)
    controllers.set_common_tracks(display_obj, chromosome_build, request.form)

    # Set up bam/cram alignments for case samples:
    controllers.set_sample_tracks(display_obj, request.form)

    # When chrom != MT, set up case-specific tracks (might be present according to the pipeline)
    if chrom != "M":
        controllers.set_case_specific_tracks(display_obj, request.form)

    display_obj["display_center_guide"] = True

    return render_template("alignviewers/igv_viewer.html", locus=locus, **display_obj)



# -*- coding: utf-8 -*-
from flask_login import current_user
from scout.server.extensions import store
from scout.constants import IGV_TRACKS, CASE_SPECIFIC_TRACKS

CUSTOM_TRACK_NAMES = ["Genes", "ClinVar", "ClinVar CNVs"]


def make_igv_tracks(name, file_list):
    """ Return a dict according to IGV track format. """
    track_list = []
    for track in file_list:
        track_list.append({"name": name, "url": track, "min": 0.0, "max": 30.0})
    return track_list


def set_common_tracks(display_obj, build, form):
    """Set up tracks common to all cases (Genes, ClinVar ClinVar CNVs) according to user preferences

    Args:
        display_obj(dict) dictionary containing all tracks info
        build(string) "37" or "38"
        form(dict) flask request form dictionary
    """
    user_obj = store.user(email=current_user.email)
    # if user settings for igv tracks exist -> use these settings, otherwise display all tracks
    custom_tracks_names = user_obj.get("igv_tracks") or CUSTOM_TRACK_NAMES

    display_obj["custom_tracks"] = []
    for track in IGV_TRACKS[build]:
        # if track is selected, add it to track display object
        if track["name"] in custom_tracks_names:
            display_obj["custom_tracks"].append(track)


def set_sample_tracks(display_obj, form):
    """Set up individual-specific alignment tracks (bam/cram files)

    Args:
        display_obj(dict) dictionary containing all tracks info
        form(dict) flask request form dictionary
    """
    samples = form.get("sample").split(",")
    sample_tracks = []
    bam_files = None
    bai_files = None
    if form.get("align") == "mt_bam":
        bam_files = form.get("mt_bam").split(",")
        bai_files = form.get("mt_bai").split(",")
    elif form.get("align") == "bam":
        bam_files = form.get("bam").split(",")
        bai_files = form.get("bai").split(",")

    counter = 0
    for sample in samples:
        # some samples might not have an associated bam file, take care if this
        if len(bam_files) > counter and bam_files[counter]:
            sample_tracks.append(
                {
                    "name": sample,
                    "url": bam_files[counter],
                    "format": bam_files[counter].split(".")[-1],  # "bam" or "cram"
                    "indexURL": bai_files[counter],
                    "height": 700,
                }
            )
        counter += 1
    display_obj["sample_tracks"] = sample_tracks


def set_case_specific_tracks(display_obj, form):
    """Set up tracks from files that might be present or not at the case level
        (rhocall files, tiddit coverage files, upd regions and sites files)
    Args:
        display_obj(dict) dictionary containing all tracks info
        form(dict) flask request form dictionary
    """
    for track, label in CASE_SPECIFIC_TRACKS.items():
        if form.get(track):
            track_info = make_igv_tracks(label, form.get(track).split(","))
            display_obj[track] = track_info



import logging

import click
from bson.json_util import dumps
from flask.cli import with_appcontext

from scout.commands.utils import builds_option
from scout.export.exon import export_exons, export_gene_exons
from scout.server.extensions import store

LOG = logging.getLogger(__name__)


@click.command("exons", short_help="Export exons")
@click.option("-b", "--build", default="37", type=click.Choice(["37", "38"]))
@click.option("-hgnc", "--hgnc-id", help="Limit search to exons of a gene", type=int)
@click.option("--json", help="Export to json format", is_flag=True)
@with_appcontext
def exons(build, hgnc_id, json):
    """Export exons to chanjo compatible .bed like format or to json"""
    LOG.info("Running scout export exons")

    if json:  # export query results to json format
        query = {"build": build}
        if hgnc_id:  # add hgnc ID to the exon query if hgnc ID is provided
            query["hgnc_id"] = hgnc_id
        result = store.exon_collection.find(query)
        click.echo(dumps(result))
        return

    # Else return them in a human-readable text table
    header = ["#Chrom\tStart\tEnd\tExonId\tTranscripts\tHgncIDs\tHgncSymbols"]

    for line in header:
        click.echo(line)

    if hgnc_id:  # print the exons of a single gene
        for exon_line in export_gene_exons(store, hgnc_id, build):
            click.echo(exon_line)
        return
    # Otherwise print all exons
    for exon_line in export_exons(store, build):
        click.echo(exon_line)



import logging
import os.path

from scout.parse.variant.managed_variant import parse_managed_variant_lines
from scout.build import build_managed_variant

LOG = logging.getLogger(__name__)


def managed_variants(store, managed_variants_query, variant_count, page=1, per_page=50):
    """Pre-process list of variants."""
    skip_count = per_page * max(page - 1, 0)
    more_variants = True if variant_count > (skip_count + per_page) else False
    managed_variants_res = managed_variants_query.skip(skip_count).limit(per_page)

    managed_variants = [managed_variant for managed_variant in managed_variants_res]

    return {"managed_variants": managed_variants, "more_variants": more_variants}


def add_managed_variant(store, add_form, institutes, current_user_id):
    """Add a managed variant."""

    managed_variant_obj = build_managed_variant(
        dict(
            chromosome=add_form["chromosome"].data,
            position=add_form["position"].data,
            end=add_form["end"].data,
            reference=add_form["reference"].data,
            alternative=add_form["alternative"].data,
            institutes=institutes,
            maintainer=[current_user_id],
            category=add_form["category"].data,
            sub_category=add_form["sub_category"].data,
            description=add_form["description"].data,
        )
    )

    return store.upsert_managed_variant(managed_variant_obj)


def upload_managed_variants(store, lines, institutes, current_user_id):
    """Add managed variants from a CSV file"""

    total_variant_lines = 0
    new_managed_variants = 0

    for managed_variant_info in parse_managed_variant_lines(lines):
        total_variant_lines += 1

        managed_variant_info.update({"maintainer": [current_user_id], "institutes": institutes})
        managed_variant_obj = build_managed_variant(managed_variant_info)

        if store.upsert_managed_variant(managed_variant_obj):
            new_managed_variants += 1

    return new_managed_variants, total_variant_lines


def modify_managed_variant(store, managed_variant_id, edit_form):
    """Modify a managed variant."""

    managed_variant = store.find_managed_variant(managed_variant_id)

    if managed_variant is None:
        return

    original_obj_id = managed_variant["_id"]

    managed_variant.update(
        {
            "chromosome": edit_form["chromosome"].data,
            "position": edit_form["position"].data,
            "end": edit_form["end"].data,
            "reference": edit_form["reference"].data,
            "alternative": edit_form["alternative"].data,
            "category": edit_form["category"].data,
            "sub_category": edit_form["sub_category"].data,
            "description": edit_form["description"].data,
        }
    )

    # new ids must be built upon update
    updated_variant = build_managed_variant(managed_variant)

    LOG.debug(
        "Updated variant has mvid %s and old id is %s.",
        updated_variant["managed_variant_id"],
        original_obj_id,
    )

    result = store.upsert_managed_variant(updated_variant, original_obj_id)

    return result


def remove_managed_variant(store, managed_variant_id):
    """Remove a managed variant."""

    removed_variant = store.delete_managed_variant(managed_variant_id)

    return removed_variant



# -*- coding: utf-8 -*-
from scout.server.extensions import store


def test_collection_stats(mock_app):
    """test the function that returns the stats for a database collection"""

    with mock_app.app_context():
        stats = store.collection_stats("case")
        assert "avgObjSize" in stats



import logging

LOG = logging.getLogger(__name__)

from scout.parse.panel import get_delimiter


def parse_managed_variant_id(
    chromosome, position, reference, alternative, category, sub_category, build="37"
):
    """Construct Managed Variant id for searching.

    Args:
        chromosome,
        position,
        reference,
        alternative,
        category,
        sub_category,
        build,

    Returns:
        managed_variant_id
    """

    return "_".join(
        [
            str(part)
            for part in (
                chromosome,
                position,
                reference,
                alternative,
                category,
                sub_category,
                build,
            )
        ]
    )


def parse_managed_variant_lines(csv_lines):
    """Parse managed variant csv lines into managed variant info dicts.

    Shares implementation structure with panel csv parsing. Could be generalised.

    Args:
        csv_lines(iterable(str))

    Returns:
        list(managed_variant_info(dict)):   A list of variant info dicts,
            possibly with the following header keys:
                chromosome,
                position,
                end,
                reference,
                alternative,
                build,
                maintainer,
                institute,
                category,
                sub_category,
                description,
    """

    managed_variant_info_dicts = []

    delimiter = "\t"

    for i, line in enumerate(csv_lines):
        line = line.rstrip()

        if line.startswith("##") or len(line) < 1:
            continue

        if line.startswith("#"):
            delimiter = get_delimiter(line)
            header = [word.lower() for word in line[1:].split(delimiter)]
            continue

        if i == 0:
            delimiter = get_delimiter(line)
            header = [word.lower() for word in line[0:].split(delimiter)]
            continue

        managed_variant_info = dict(zip(header, line.split(delimiter)))

        # There are cases when excel exports empty lines
        if not any(managed_variant_info.values()):
            continue

        managed_variant_info_dicts.append(managed_variant_info)

    return managed_variant_info_dicts



"""Code for updating information on individuals"""
from pathlib import Path

import click
from scout.server.extensions import store

UPDATE_KEYS = [
    "bam_file",
    "mt_bam",
    "vcf2cytosure",
    "rhocall_bed",
    "rhocall_wig",
    "tiddit_coverage_wig",
    "upd_regions_bed",
    "upd_sites_bed",
]


@click.command()
@click.option("--case-id", "-c", required=True, help="Case id")
@click.option("--ind", "-n", help="Individual display name")
@click.argument("key", required=False)
@click.argument("value", required=False)
def individual(case_id, ind, key, value):
    """Update information on individual level in Scout"""

    case_obj = store.case(case_id)
    if not case_obj:
        click.echo(f"Could not find case {case_id}")
        return
    individuals = {ind_info["display_name"]: ind_info for ind_info in case_obj["individuals"]}
    # If ind name is empty, print available individual names for this case to help the user to build the command
    if ind is None:
        click.echo(
            f"Please specify individual name with '-n' option. Available individuals for this case:{list(individuals.keys())}"
        )
        return
    if ind not in individuals:
        click.echo(
            f"Could not find individual '{ind}' in case individuals. Available individuals for this case: {list(individuals.keys)}"
        )
        return
    # If key is null or non-valid, print a list of all the keys that can be updated using this function
    if key is None or not key in UPDATE_KEYS:
        click.echo(f"Please specify a valid key to update. Valid keys:{ UPDATE_KEYS }")
        return
    if value is None:
        click.echo(f"Please specify a file path for key {key}")
        return
    file_path = Path(value)
    # If file is not found on the server, ask if user wants to update the key anyway
    if file_path.exists() is False:
        click.confirm(
            "The provided path was not found on the server, update key anyway?", abort=True
        )
    # perform the update
    for ind_obj in case_obj["individuals"]:
        if ind_obj["display_name"] == ind:
            ind_obj[key] = value

    store.update_case(case_obj)



import logging

import click
from flask import url_for, current_app
from flask.cli import with_appcontext
from scout.server.extensions import store

LOG = logging.getLogger(__name__)

BYTES_IN_ONE_GIGABYTE = 1073741824  # (1024*1024*1024)
DELETE_VARIANTS_HEADER = [
    "Case n.",
    "Ncases",
    "Institute",
    "Case name",
    "Case ID",
    "Case track",
    "Analysis date",
    "Status",
    "Research",
    "Total variants",
    "Removed variants",
]
CASE_STATUS = ["solved", "archived", "migrated", "active", "inactive", "prioritized"]


@click.command("variants", short_help="Delete variants for one or more cases")
@click.option("-u", "--user", help="User running this command (email)", required=True)
@click.option("-c", "--case-id", help="Case id")
@click.option(
    "--status",
    type=click.Choice(CASE_STATUS),
    multiple=True,
    default=[],
    help="Restrict to cases with specified status",
)
@click.option("--older-than", type=click.INT, default=0, help="Older than (months)")
@click.option(
    "--analysis-type",
    type=click.Choice(["wgs", "wes", "panel"]),
    multiple=True,
    default=["wgs"],
    help="Type of analysis",
)
@click.option("--rank-threshold", type=click.INT, default=5, help="With rank threshold lower than")
@click.option("--variants-threshold", type=click.INT, help="With more variants than")
@click.option(
    "--dry-run",
    is_flag=True,
    help="Perform a simulation without removing any variant",
)
@with_appcontext
def variants(
    user: str,
    case_id: str,
    status: list,
    older_than: int,
    analysis_type: list,
    rank_threshold: int,
    variants_threshold: int,
    dry_run: bool,
) -> None:
    """Delete variants for one or more cases"""

    user_obj = store.user(user)
    if user_obj is None:
        click.echo(f"Could not find a user with email '{user}' in database")
        return

    total_deleted = 0
    items_name = "deleted variants"
    if dry_run:
        click.echo("--------------- DRY RUN COMMAND ---------------")
        items_name = "estimated deleted variants"
    else:
        click.confirm("Variants are going to be deleted from database. Continue?", abort=True)

    case_query = store.build_case_query(case_id, status, older_than, analysis_type)
    # Estimate the average size of a variant document in database
    avg_var_size = store.collection_stats("variant").get("avgObjSize", 0)  # in bytes

    # Get all cases where case_query applies
    n_cases = store.case_collection.count_documents(case_query)
    cases = store.cases(query=case_query)
    filters = (
        f"Rank-score threshold:{rank_threshold}, case n. variants threshold:{variants_threshold}."
    )
    click.echo("\t".join(DELETE_VARIANTS_HEADER))
    for nr, case in enumerate(cases, 1):
        case_id = case["_id"]
        case_n_variants = store.variant_collection.count_documents({"case_id": case_id})
        # Skip case if user provided a number of variants to keep and this number is less than total number of case variants
        if variants_threshold and case_n_variants < variants_threshold:
            continue
        # Get evaluated variants for the case that haven't been dismissed
        case_evaluated = store.evaluated_variants(case_id=case_id)
        evaluated_not_dismissed = [
            variant["_id"] for variant in case_evaluated if "dismiss_variant" not in variant
        ]
        # Do not remove variants that are either pinned, causative or evaluated not dismissed
        variants_to_keep = (
            case.get("suspects", []) + case.get("causatives", []) + evaluated_not_dismissed or []
        )
        variants_query = store.delete_variants_query(case_id, variants_to_keep, rank_threshold)

        if dry_run:
            # Just print how many variants would be removed for this case
            remove_n_variants = store.variant_collection.count_documents(variants_query)
            total_deleted += remove_n_variants
            click.echo(
                "\t".join(
                    [
                        str(nr),
                        str(n_cases),
                        case["owner"],
                        case["display_name"],
                        case_id,
                        case.get("track", ""),
                        str(case["analysis_date"]),
                        case.get("status", ""),
                        str(case.get("is_research", "")),
                        str(case_n_variants),
                        str(remove_n_variants),
                    ]
                )
            )
            continue

        # delete variants specified by variants_query
        result = store.variant_collection.delete_many(variants_query)
        click.echo(f"Deleted {result.deleted_count} / {case_n_variants} total variants")
        total_deleted += result.deleted_count
        click.echo(
            "\t".join(
                [
                    str(nr),
                    str(n_cases),
                    case["owner"],
                    case["display_name"],
                    case_id,
                    case.get("track", ""),
                    str(case["analysis_date"]),
                    case.get("status", ""),
                    str(case.get("is_research", "")),
                    str(case_n_variants),
                    str(result.deleted_count),
                ]
            )
        )

        # Create event in database
        institute_obj = store.institute(case["owner"])
        with current_app.test_request_context("/cases"):
            url = url_for(
                "cases.case", institute_id=institute_obj["_id"], case_name=case["display_name"]
            )
            store.remove_variants_event(
                institute=institute_obj, case=case, user=user_obj, link=url, content=filters
            )

    click.echo(f"Total {items_name}: {total_deleted}")
    click.echo(
        f"Estimated space freed (GB): {round((total_deleted * avg_var_size) / BYTES_IN_ONE_GIGABYTE, 4)}"
    )


@click.command("panel", short_help="Delete a gene panel")
@click.option("--panel-id", help="The panel identifier name", required=True)
@click.option("-v", "--version", type=float)
@with_appcontext
def panel(panel_id, version):
    """Delete a version of a gene panel or all versions of a gene panel"""
    LOG.info("Running scout delete panel")
    adapter = store

    res = adapter.gene_panels(panel_id=panel_id, version=version)
    panel_objs = [panel_obj for panel_obj in res]
    if len(panel_objs) == 0:
        LOG.info("No panels found")

    for panel_obj in panel_objs:
        adapter.delete_panel(panel_obj)


@click.command("index", short_help="Delete all indexes")
@with_appcontext
def index():
    """Delete all indexes in the database"""
    LOG.info("Running scout delete index")
    adapter = store

    for collection in adapter.db.list_collection_names():
        adapter.db[collection].drop_indexes()
    LOG.info("All indexes deleted")


@click.command("user", short_help="Delete a user")
@click.option("-m", "--mail", required=True)
@with_appcontext
def user(mail):
    """Delete a user from the database"""
    LOG.info("Running scout delete user")
    adapter = store
    user_obj = adapter.user(mail)
    if not user_obj:
        LOG.warning("User {0} could not be found in database".format(mail))
        return

    result = adapter.delete_user(mail)
    if result.deleted_count == 0:
        return
    click.echo(f"User was correctly removed from database.")
    # remove this user as assignee from any case where it is found
    assigned_cases = adapter.cases(assignee=mail)
    updated_cases = 0
    with current_app.test_request_context("/cases"):
        for case_obj in assigned_cases:
            institute_obj = adapter.institute(case_obj["owner"])
            link = url_for(
                "cases.case", institute_id=institute_obj["_id"], case_name=case_obj["display_name"]
            )
            inactivate_case = case_obj.get("status", "active") == "active" and case_obj[
                "assignees"
            ] == [mail]
            if adapter.unassign(institute_obj, case_obj, user_obj, link, inactivate_case):
                updated_cases += 1
    click.echo(f"User was removed as assignee from {updated_cases} case(s).")


@click.command("genes", short_help="Delete genes")
@click.option("-b", "build", type=click.Choice(["37", "38"]))
@with_appcontext
def genes(build):
    """Delete all genes in the database"""
    LOG.info("Running scout delete genes")
    adapter = store

    if build:
        LOG.info("Dropping genes collection for build: %s", build)
    else:
        LOG.info("Dropping all genes")
    adapter.drop_genes(build=build)


@click.command("exons", short_help="Delete exons")
@click.option("-b", "build", type=click.Choice(["37", "38"]))
@with_appcontext
def exons(build):
    """Delete all exons in the database"""
    LOG.info("Running scout delete exons")
    adapter = store

    adapter.drop_exons(build)


@click.command("case", short_help="Delete a case")
@click.option("-i", "--institute", help="institute id of related cases")
@click.option("-c", "--case-id")
@click.option("-d", "--display-name")
@with_appcontext
def case(institute, case_id, display_name):
    """Delete a case and it's variants from the database"""
    adapter = store
    case_obj = None

    if not (case_id or display_name):
        click.echo("Please specify what case to delete")
        raise click.Abort()

    if display_name:
        if not institute:
            click.echo(
                "Please specify the owner of the case that should be "
                "deleted with flag '-i/--institute'."
            )
            raise click.Abort()
        case_obj = adapter.case(display_name=display_name, institute_id=institute)
    else:
        case_obj = adapter.case(case_id=case_id)

    if not case_obj:
        click.echo("Coudn't find any case in database matching the provided parameters.")
        raise click.Abort()

    LOG.info("Running deleting case {0}".format(case_id))
    case = adapter.delete_case(case_id=case_obj["_id"])

    adapter.delete_variants(case_id=case_obj["_id"], variant_type="clinical")
    adapter.delete_variants(case_id=case_obj["_id"], variant_type="research")


# @click.command('diseases', short_help='Display all diseases')
# @click.pass_context
# def diseases(context):
#     """Show all diseases in the database"""
#     LOG.info("Running scout view diseases")
#     adapter = context.obj['adapter']
#
#     click.echo("Disease")
#     for disease_obj in adapter.disease_terms():
#         click.echo("{0}:{1}".format(
#             disease_obj['source'],
#             disease_obj['disease_id'],
#         ))
#
# @click.command('hpo', short_help='Display all hpo terms')
# @click.pass_context
# def hpo(context):
#     """Show all hpo terms in the database"""
#     LOG.info("Running scout view hpo")
#     adapter = context.obj['adapter']
#
#     click.echo("hpo_id\tdescription")
#     for hpo_obj in adapter.hpo_terms():
#         click.echo("{0}\t{1}".format(
#             hpo_obj.hpo_id,
#             hpo_obj.description,
#         ))


@click.group()
def delete():
    """
    Delete objects from the database.
    """
    pass


delete.add_command(panel)
delete.add_command(genes)
delete.add_command(case)
delete.add_command(user)
delete.add_command(index)
delete.add_command(exons)
delete.add_command(variants)



# -*- coding: utf-8 -*-
from flask import url_for, current_app
from flask_login import current_user
from urllib.parse import urlencode
from scout.server.extensions import store


def test_acmg(app):

    # GIVEN an initialized app
    with app.test_client() as client:

        # the acmg endpoint endpoint should return an acmg json file
        resp = client.get("/api/v1/acmg")
        assert resp.status_code == 200
        assert resp.data


def test_variant(app, institute_obj, case_obj, variant_obj):
    # GIVEN an initialized app
    # GIVEN a valid user and institute

    with app.test_client() as client:
        # GIVEN that the user could be logged in
        resp = client.get(url_for("auto_login"))
        assert resp.status_code == 200

        # WHEN sending a request (GET) to the variant page
        resp = client.get(
            url_for(
                "variant.variant",
                institute_id=institute_obj["internal_id"],
                case_name=case_obj["display_name"],
                variant_id=variant_obj["_id"],
            )
        )
        # THEN it should return a page
        assert resp.status_code == 200


def test_sv_variant(app, institute_obj, case_obj, variant_obj):
    # GIVEN an initialized app
    # GIVEN a valid user and institute

    with app.test_client() as client:
        # GIVEN that the user could be logged in
        resp = client.get(url_for("auto_login"))
        assert resp.status_code == 200

        # WHEN sending a request (GET) to the sv_variant page
        resp = client.get(
            url_for(
                "variant.sv_variant",
                institute_id=institute_obj["internal_id"],
                case_name=case_obj["display_name"],
                variant_id=variant_obj["_id"],
            )
        )
        # THEN it should return a page
        assert resp.status_code == 200


def test_variant_update_manual_rank(app, case_obj, variant_obj, institute_obj):
    # GIVEN an initialized app
    # GIVEN a valid user and institute

    with app.test_client() as client:
        # GIVEN that the user could be logged in
        resp = client.get(url_for("auto_login"))
        assert resp.status_code == 200

        # When a manual Rank is assigned to the variant via POST request
        data = urlencode({"manual_rank": "7"})  # pathogenic
        resp = client.post(
            url_for(
                "variant.variant_update",
                institute_id=institute_obj["internal_id"],
                case_name=case_obj["display_name"],
                variant_id=variant_obj["_id"],
                data=data,
                content_type="application/x-www-form-urlencoded",
            )
        )
        # THEN request should be a redirection
        assert resp.status_code == 302


def test_edit_variants_comments(app, institute_obj, case_obj, user_obj, variant_obj):
    """Test the functionality to modify a variants comment"""

    # GIVEN an initialized app
    with app.test_client() as client:
        # GIVEN that the user could be logged in
        resp = client.get(url_for("auto_login"))

        ## GIVEN a variant with a specific comment
        store.create_event(
            institute=institute_obj,
            case=case_obj,
            user=user_obj,
            variant=variant_obj,
            link="a link",
            category="variant",
            verb="comment",
            content="a specific comment",
            subject=variant_obj["display_name"],
            level="specific",
        )
        comment = store.event_collection.find_one({"verb": "comment"})
        assert comment
        assert comment["level"] == "specific"

        # WHEN a user updates the comment via the modal form
        form_data = {
            "event_id": comment["_id"],
            "updatedContent": "a global comment",
            "edit": "",
            "level": "global",
        }
        resp = client.post(
            url_for(
                "cases.events",
                institute_id=institute_obj["_id"],
                case_name=case_obj["display_name"],
                event_id=comment["_id"],
            ),
            data=form_data,
        )
        # THEN it should redirect to case page
        assert resp.status_code == 302

        # And the comment should be updated
        updated_comment = store.event_collection.find_one()
        assert updated_comment["content"] == "a global comment"
        assert updated_comment["level"] == "global"


def test_variant_update_cancer_tier(app, case_obj, variant_obj, institute_obj):
    # GIVEN an initialized app
    # GIVEN a valid user and institute

    with app.test_client() as client:
        # GIVEN that the user could be logged in
        resp = client.get(url_for("auto_login"))
        assert resp.status_code == 200

        # When a cancer tier is assigned to the variant via POST request
        data = urlencode({"cancer_tier": "2C"})  # pathogenic
        resp = client.post(
            url_for(
                "variant.variant_update",
                institute_id=institute_obj["internal_id"],
                case_name=case_obj["display_name"],
                variant_id=variant_obj["_id"],  # let's assume this is a cancer variant
                data=data,
                content_type="application/x-www-form-urlencoded",
            )
        )
        # THEN request should be a redirection
        assert resp.status_code == 302


def test_clinvar(app, case_obj, variant_obj, institute_obj):
    # GIVEN an initialized app
    # GIVEN a valid user and institute

    with app.test_client() as client:
        # GIVEN that the user could be logged in
        resp = client.get(url_for("auto_login"))
        assert resp.status_code == 200

        # WHEN the page is accessed via GET request
        resp = client.get(
            url_for(
                "variant.clinvar",
                institute_id=institute_obj["internal_id"],
                case_name=case_obj["display_name"],
                variant_id=variant_obj["_id"],
            )
        )
        # Then it should return a page
        assert resp.status_code == 200

        # WHEN instead the page is requested via POST method
        # using a test clinvar form:
        data = urlencode(
            {
                "main_var": variant_obj["_id"],
                "@".join(["local_id", variant_obj["_id"]]): variant_obj["_id"],
                "@".join(["chromosome", variant_obj["_id"]]): variant_obj["chromosome"],
                "@".join(["start", variant_obj["_id"]]): variant_obj["position"],
                "@".join(["stop", variant_obj["_id"]]): variant_obj["end"],
                "@".join(["condition_id_value", variant_obj["_id"]]): [
                    "HPO_HP:0001250",
                    "OMIM_145590",
                ],
                "@".join(["clin_features", variant_obj["_id"]]): ["HPO_HP:0001507"],
            }
        )
        resp = client.post(
            url_for(
                "variant.clinvar",
                institute_id=institute_obj["internal_id"],
                case_name=case_obj["display_name"],
                variant_id=variant_obj["_id"],
            ),
            data=data,
            content_type="application/x-www-form-urlencoded",
        )
        # THEN the request status should be a redirect
        assert resp.status_code == 302


def test_update_tracks_settings(app, user_obj):
    """Test the endpoint that updates the IGV track preferences for a user"""

    preferred_tracks = ["Genes", "ClinVar"]
    # GIVEN an initialized app
    with app.test_client() as client:
        # GIVEN that the user could be logged in
        resp = client.get(url_for("auto_login"))

        # GIVEN that the user wants to see only Genes and ClinVar SNVs tracks
        form_data = {
            "user_tracks": preferred_tracks,
        }

        # WHEN sending a POST request to the update
        resp = client.post(
            url_for(
                "variant.update_tracks_settings",
            ),
            data=form_data,
        )

        # THEN the user object in the database should be updated with the right track info
        user_obj = store.user(email=user_obj["email"])
        for track in preferred_tracks:
            assert track in preferred_tracks
